#ifdef _WIN32
#include <windows.h>
#include <stdio.h>
#else
#include <tap.h>
#endif
#include "hash.h"


typedef struct {
   int hash_length;         //bytes
   int hash_elements;
   type_hash_element **hash_table;
} type_hash;


//*
//* The FCS-32 generator polynomial: x**0 + x**1 + x**2 + x**4 + x**5
//*                      + x**7 + x**8 + x**10 + x**11 + x**12 + x**16
//*                      + x**22 + x**23 + x**26 + x**32.
//*
#define P32             0xedb88320  // the polymonial
#define CRC_POLYNOMIAL  0x04c11db7
//*
//* FCS lookup table as calculated by the table generator.
//*
static dword fcstab8[0x100];

static dword bitrevN(int n, dword data)
{
    int i;
    dword result;

    result = 0;
    for (i=0; i<n; i++) {
        result += ((data >> (n-1-i)) & 1) << (i);
    }
    return result;
}

void init_crc(void)
{
    int b;
    int b2;
    dword v;
    int i;

    for (b=0; b<(1<<8); b++) {
        b2 = bitrevN(8,b);
        v = (dword)b;
        for(i=0; i<8; i++) {
            v = (v >> 1) ^ ((v & 1) ? P32 : 0);
        }

        fcstab8[b2] = bitrevN(32,v);
    }
}

// 32 bit crc, 8 bit data, 8 bits at a time
// lookup table = 2^8=256 words
dword crc8(dword crc, byte *data, int len)
{
    int i;
    int index;
    dword dataout32;

    for(i=0; i<len; i++) {
        index = ((crc>>24) ^ (data[i])) & 0xff;
        dataout32 = fcstab8[index];
        crc = (crc << 8) ^ dataout32;
    }

    return crc;
}

// Shrink HASHLENGTH-bit number to an log(HASHELEMENTS)-bit number (HASHELEMENTS)
static dword hash(type_hash *handle, byte *n)
{
    dword crc;

    crc = crc8(0xffffffff, n, handle->hash_length);
    return crc % handle->hash_elements;
}

void *open_hash(int hash_length, int hash_elements)
{
    type_hash *handle;

    handle = (type_hash *)malloc(sizeof(type_hash));
    handle->hash_length = hash_length;
    handle->hash_elements = hash_elements;
    handle->hash_table = (type_hash_element **)malloc(handle->hash_elements * sizeof(type_hash_element *));
    memset(handle->hash_table, 0, handle->hash_elements * sizeof(type_hash_element *));
    return handle;
}

static void free_hash(type_hash_element *curhash)
{
    if (curhash->next) {
        free_hash(curhash->next);
    }
    free(curhash);
}

int close_hash(void *handle)
{
    int i;
    type_hash *h;

    if (!handle) return -1;
    h = (type_hash *)handle;

    //free up any outstanding hashes
    for (i=0; i<h->hash_elements; i++) {
        if (h->hash_table[i]) {
            free_hash(h->hash_table[i]);
        }
    }

    free(h->hash_table);
    free(h);
    return 0;
}

static type_hash_element *new_hash(type_hash *handle, byte *key, dword data)
{
    type_hash_element *newhash;

    newhash = (type_hash_element*)malloc(sizeof(type_hash_element) - 4 + handle->hash_length);
    memcpy(newhash->key, key, handle->hash_length);
    newhash->data = data;
    newhash->next = NULL;
    return newhash;
}

//should addhash check for existing key, or just blindly add a (possibly duplicate) new one
//#define NO_DUPLICATES
type_hash_element *addhash(void *handle, byte *key, dword data)
{
    dword hashval;
    type_hash *h;
    type_hash_element **hashtab;
    type_hash_element *curhash;
    type_hash_element *newhash;

    hashval = hash(handle, key);

    h = (type_hash *)handle;
    hashtab = ((type_hash *)handle)->hash_table;

    curhash = hashtab[hashval];
    if (curhash) {
#ifdef NO_DUPLICATES
        if ((memcmp(curhash->key, key, h->hash_length) == 0) {
            return curhash;
        }
#endif
        while (curhash->next) {
            curhash = curhash->next;
#ifdef NO_DUPLICATES
            if ((memcmp(curhash->key, key, h->hash_length) == 0) {
                return curhash;
            }
#endif
        }
        newhash = new_hash(handle, key, data);
        curhash->next  = newhash;
    }
    else {
        newhash = new_hash(handle, key, data);
        hashtab[hashval] = newhash;
    }
    return newhash;
}

type_hash_element *findhash(void *handle, byte *key)
{
    dword hashval;
    type_hash *h;
    type_hash_element **hashtab;
    type_hash_element *curhash;

    hashval = hash(handle, key);

    h = (type_hash *)handle;
    hashtab = ((type_hash *)handle)->hash_table;
    if (hashtab == NULL) {
//        TAP_Print("\nhashtab=%x", hashtab);
        return NULL;
    }
    if (hashtab[hashval] == NULL) {
        return NULL;
    }

    curhash = hashtab[hashval];
    if (memcmp(curhash->key, key, h->hash_length) == 0) {
        return curhash;
    }

    while (curhash->next) {
        curhash = curhash->next;
        if (memcmp(curhash->key, key, h->hash_length) == 0) {
            return curhash;
        }
    }
    return NULL;
}

int delhash(void *handle, byte *key)
{
    dword hashval;
    type_hash *h;
    type_hash_element **hashtab;
    type_hash_element* lasthash;
    type_hash_element* curhash;

    hashval = hash(handle, key);

    h = (type_hash *)handle;
    hashtab = ((type_hash *)handle)->hash_table;
    if (hashtab[hashval] == NULL) {
        return 0;
    }

    curhash = hashtab[hashval];
    if (memcmp(curhash->key, key, h->hash_length) == 0) {
        hashtab[hashval] = curhash->next;
        free(curhash);
        return 1;
    }
    while (curhash->next) {
        lasthash = curhash;
        curhash = curhash->next;
        if (memcmp(curhash->key, key, h->hash_length) == 0) {
            lasthash->next = curhash->next;
            free(curhash);
            return 1;
        }
    }
    return 0;
}


//============================================================================
#ifdef MAIN
#define HASHLENGTH              6
#define HASHELEMENTS            5000

typedef struct{
    word original_network_id;
    word transport_stream_id;
    word event_id;
    word service_id;
}type_events;

int main()
{
    void *hash_handle;
    type_hash_element *hash_ptr;
    dword data;
    type_events event_hash = {0};

    init_crc();
    hash_handle = open_hash(HASHLENGTH, HASHELEMENTS);
printf("\nhash_handle=%x",hash_handle);
    event_hash.original_network_id = 0x1234;
    event_hash.transport_stream_id = 0x5678;
    event_hash.service_id          = 0x9abc;
    event_hash.event_id            = 0xdef0;
    printf("\nh(%I64x)=%x", *(dword *)&event_hash, hash(hash_handle, (byte *)&event_hash));


printf("\n");
    data = 1;
    hash_ptr = addhash(hash_handle, (byte *)&event_hash, data);
    printf("\nhash_ptr = %08x", hash_ptr);
    printf("\nkey  = %I64x", *(ulong64 *)hash_ptr->key);
    printf("\ndata = %x", hash_ptr->data);

printf("\n");
    data = 2;
    hash_ptr = addhash(hash_handle, (byte *)&event_hash, data);
    printf("\nhash_ptr = %08x", hash_ptr);
    printf("\nkey  = %I64x", *(ulong64 *)hash_ptr->key);
    printf("\ndata = %x", hash_ptr->data);


printf("\n");
    hash_ptr = findhash(hash_handle, (byte *)&event_hash);
    printf("\nfindhash(%I64x)", *(ulong64 *)&event_hash);
    printf("\nhash_ptr = %08x", hash_ptr);
    printf("\nkey  = %I64x", *(ulong64 *)hash_ptr->key);
    printf("\ndata = %x", hash_ptr->data);

//    delhash(hash_handle, *(dword *)&event_hash);
printf("\n");
    event_hash.service_id          = 0x9999;
    hash_ptr = findhash(hash_handle, (byte *)&event_hash);
    printf("\nfindhash(%I64x)", *(ulong64 *)&event_hash);
    printf("\nhash_ptr = %08x", hash_ptr);
    printf("\nkey  = %I64x", *(ulong64 *)hash_ptr->key);
    printf("\ndata = %x", hash_ptr->data);

    printf("\nh(%I64x)=%x", *(ulong64 *)&event_hash, hash(hash_handle, (byte *)&event_hash));

    delhash(hash_handle, (byte *)&event_hash);
    close_hash(hash_handle);

    return 0;
}
#endif
//============================================================================
